""" base classes to describe python code """
__revision__ = '$Id: objects.py,v 1.3 2003/02/18 10:11:15 syt Exp $'

import re
from logilab.common.visitor import VisitedMixIn

# regexps to try to describe python code as if it was java code ;)
#          handle _ visibility
#                 _ class / abstract class / interface differentiation
#                 _ final
#
PROTECTED = re.compile('^_[A-Za-z]+(_*[A-Za-z]+)*_?$')
PRIVATE = re.compile('^__[A-Za-z]+(_*[A-Za-z]+)*_?$')
SPECIAL = re.compile('^__[A-Za-z]+__$')
FINAL = re.compile('^[A-Z_]*$')
ABSTRACT = re.compile('^.*Abstract.*')
INTERFACE = re.compile('^.*Interface.*')
EXCEPTION = re.compile('^.*(exception|error).*', re.I)


# mixin to handle locals and globals dict ######################################

class FrameMixIn:
    """
    Mixin to handle a dict of defined object in a context
    
    /!\ /!\ it should be the leftmost ancestor
    """
    
    def __init__(self):
        # imports modules/objects
        self.imports = []
        # raised exceptions
        self.raises = []
        # dictionnary equivalent to global of this node
        self._defs= {}

    def add_import(self, fullname, asname):
        """
        add an import statement to the context
        """
        self._defs[asname] = fullname
        self.imports.append((fullname, asname))
        
    def add_raised_exception(self, exception):
        """
        add a raise statement to the context
        """
        if not exception in self.raises:
            self.raises.append(exception)

    def get_ref(self, name):
        """
        asking for a reference to an object named 'name'
        if the object is not found in this context, ask to my parent
        """
        mychance = self._defs.get(name)
        if mychance and mychance != name:
            return mychance
        elif self.parent:
            subname = '.'.join(self.fullname().split('.')[:-1])
            if subname:
                other = self.parent.get_ref('%s.%s'%(subname,name))
                if other:
                    return other
            return self.parent.get_ref(name)
        
    def appendChild(self, child):
        """
        overload AbstractBase.appendChild
        """
        if AbstractBase.appendChild(self, child):
            self._defs[child.name] = child
            return 1


# base abstract classes ######################################################

class AbstractBase(VisitedMixIn):
    """
    base class, all others inherit from this one
    """
    SP = ['children', 'id', 'parent', 'name']

    def __init__(self, name, id):
        self.id = id
        self.name = name
        self.parent = None
        self.children = []
        self.doc = ''
        self._children_d = {}
        if PROTECTED.match(name):
            self.visibility = 'protected'
        elif PRIVATE.match(name):
            self.visibility = 'private'
        else:
            self.visibility = 'public'
        if SPECIAL.match(name):
            self.special = 1
        else:
            self.special = 0
            
    def fullname(self):
        """
        for instance return 'module_name.class_name'
        """
        if self.parent:
            pname = self.parent.fullname()
            if pname:
                return '%s.%s'% (pname, self.name)
        return self.name

    def root(self):
        """
        return project node
        """
        if self.parent is None:
            return self
        return self.parent.root()

    def get_module(self):
        """
        get the first ancestor which is a module
        """
        if self.TYPE == 'module':
            return self
        if self.parent:
            return self.parent.get_module()
        
    def get_class(self):
        """
        get the first ancestor which is a class
        """
        if self.TYPE == 'class':
            return self
        if self.parent:
            return self.parent.get_class()
        
    def appendChild(self, child):
        """
        add a new child (unique name)
        """
        if not self._children_d.has_key(child.name):
            self._children_d[child.name] = child
            self.children.append(child)
            child.parent = self
            return 1
        return 0
    
    def __repr__(self):
        """
        short representation (no recursion)
        """
        return '<%s name="%s" id="%d">' % (self.TYPE, self.name, self.id)

    def __str__(self, i=0):
        """
        xml like representation
        """
        s = '%s<%s name="%s" id="%d"'%(i*'  ', self.TYPE,
                                          self.name, self.id)
        for a in self.__dict__.keys():
            if not a in self.SP:
                val = getattr(self, a)
                if val:
                    s = '%s\n%s%s="%s"'%(s,(i+1)*'  ', a, val)
        if not self.children:
            return '%s/>\n'%s
        else:
            s = '%s>\n'%s
            for c in self.children:
                s = '%s%s'%(s, c.__str__(i+1))
            return '%s%s</%s>\n'%(s, i*'  ', self.TYPE)

        
class AbstractFinal(AbstractBase):
    """
    base class for element finalizable
    """

    def __init__(self, name, id):
        AbstractBase.__init__(self, name, id)
        if FINAL.match(name):
            self.final = 'true'
        else:
            self.final = 'false'


class AbstractAttr(AbstractFinal):
    """
    base class for class child (attributes and functions)
    """

    def __init__(self, name, id, type=None):
        AbstractFinal.__init__(self, name, id)
        self.type = type # instance/class

    
# classes used to describe a project implementation ############################
            
class Project(FrameMixIn, AbstractBase):
    """
    class representing a project this is the root of the hierarchy tree
    a project can only contains modules
    """
    
    TYPE = 'project'

    def __init__(self, name, id):
        FrameMixIn.__init__(self)
        AbstractBase.__init__(self, name, id)

    def get_classes(self, name, module=None):
        """
        return the list of classes named 'name'
        (you can optionaly specify a module name)
        """
        r = []
        if module:
            m = self.get_module(module)
            if m:
                c = module.get_class(name)
                if c: r.append(c)
        else:
            for module in self.children:
                c = module.get_class(name)
                if c: r.append(c)
        return r

    def get_module(self, name=''):
        """
        overload AbstractBase.get_module
        """
        if name:
            for module in self.children:
                if module.name == name:
                    return module
                
    def complete(self, filter):
        """
        call visitors to resolve links between node of the tree
        """
        from logilab.pyreverse.visitors import inspector
        visitor = inspector.NamespaceInspector(filter)
        visitor.visit(self)
        visitor = inspector.AttributeInspector(filter)
        visitor.visit(self)
                    
    def fullname(self):
        """
        overload AbstractBase.fullname
        """
        return ''
   

class Module(FrameMixIn, AbstractBase):
    """
    class representing a module
    a module can contain classes, attributes and functions
    """

    TYPE = 'module'
    
    def __init__(self, name, id):
        FrameMixIn.__init__(self)
        AbstractBase.__init__(self, name, id)

    def get_deps(self):
        """
        return module dependencies
        """
        # look for cache
        if hasattr(self, '_deps'):
            return self._deps
        # nevermind, calculate it 
        deps = {}
        for defi in self._defs.values():
            try:
                m = defi.get_module()
                if m.name != self.name:
                    deps[m] = 1
            except:
                pass
        self._deps = deps.keys()
        return self._deps

    def fullname(self):
        """
        overload AbstractBase.fullname
        """
        name = self.name.split('.')
        if name[-1] == '__init__':
            return '.'.join(name[:-1])
        return '.'.join(name)

    def get_class(self, name):
        """
        return the class named 'name' if present
        """
        for child in self.children:
            if child.TYPE == 'class' and child.name == name:
                return child
        return None


class Klass(FrameMixIn, AbstractFinal):
    """
    class representing a class
    a class can contain classes, attributes and functions
    """

    TYPE = 'class'
    
    def __init__(self, name, id):
        FrameMixIn.__init__(self)
        AbstractFinal.__init__(self, name, id)
        self.parents, self.subclasses = [], []
        if INTERFACE.match(name):
            self.type = 'interface' 
        else: self.type = 'class'
        if ABSTRACT.match(name):
            self.abstract = 'true'
        else: self.abstract= 'false'
        if EXCEPTION.match(name):
            self.signal = 'true'
        else: self.signal = 'false'

                
class Function(FrameMixIn, AbstractAttr):
    """
    class representing a function
    a function can contains classes, attributes and functions
    """

    TYPE = 'function'        

    def __init__(self, name, id, type=None):
        FrameMixIn.__init__(self)
        AbstractAttr.__init__(self, name, id, type)

    def fullname(self):
        """
        overload AbstractBase.fullname
        """
        name = self.name
        pname = self.parent.fullname()
        if pname:
            return '%s.%s'% (pname, name)
        return name


class Attribute(AbstractAttr):
    """
    class representing an attribute
    an attribute can't contains any element
    """

    TYPE = 'attribute'


class Parameter(AbstractFinal):
    """
    class representing an parameter
    an parameter can't contains any element
    """

    TYPE = 'parameter'
    
    def __init__(self, name, id, type):
        AbstractFinal.__init__(self, name, id)
        self.type = type ## in/out
        
        
