# Copyright (c) 2000-2002 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
"""

from logilab.common.visitor import Visitor
from logilab.common.tree import PrefixedDepthFirstIterator
from logilab.pyreverse.base.lib import BasePyReverseVisitor
from logilab.pyreverse.base.objects import AbstractBase


__revision__ = "$Id: stats.py,v 1.3 2003/02/18 11:17:36 syt Exp $"

from sys import prefix, version
from os.path import join
STD_LIB_DIR = join(prefix, 'lib', 'python%s'%version[:3])
PREFIX_SIZE = len(STD_LIB_DIR)

def is_standard_module(modname):
    try:
        m = __import__(modname)
    except:
        return 0
    try:
        if (m.__file__[:PREFIX_SIZE] == STD_LIB_DIR
            and m.__file__[PREFIX_SIZE:PREFIX_SIZE+13] != 'site-packages'):
            return 1
    except:
        return 1
    return 0

class StatsVisitor(Visitor, BasePyReverseVisitor):
    """
    visitor which calculates statistics from a project description
    """

    def __init__(self, data=None):
        Visitor.__init__(self, PrefixedDepthFirstIterator)
        if data is None:
            self.data = {'stat' : {},
                         'unresolved' : {},
                         'import*' : {},
                         'nice_stats': {},
                         'bad_default_value': [],
                         'undocummented': [],
                         'cycle' : []
                         }
        else:
            self.data = data
        return data

    def close_visit(self, result):
        """
        calculate nicer stats
        """
        data = self.data
        # percentage of modules documented
        total, documented = data['stat']['module']
        percent = float(documented*100) / total
        data['nice_stats']['module'] = (total, documented, percent)
        # percentage of classes documented
        try:
            total, documented = data['stat']['class']
            percent = float(documented*100) / total
            data['nice_stats']['class'] = (total, documented, percent)
        except KeyError:
            data['nice_stats']['class'] = (0, 0, 0)
        # percentage of functions documented
        try:
            total, documented = data['stat']['function']
            percent = float(documented*100) / total
            data['nice_stats']['function'] = (total, documented, percent)
        except KeyError:
            data['nice_stats']['function'] = (0, 0, 0)
        # remove duplicate from cycle
        done = []
        for _from, to in data['cycle']:
            if not (_from, to) in done:
                done.append((to, _from))
        data['cycle'] = done
        return self.data
    
    def visit_module(self, node):
        self._stat(node)
        data = self.data
        for dep in node.get_deps():
            if node in dep.get_deps():
                data['cycle'].append((node, dep))
    
    def visit_class(self, node):
        self._stat(node)
    
    def visit_function(self, node):
        self._stat(node)
    
    def visit_attribute(self, node):
        pass
        #self._stat_default_value(node)
            
    def visit_parameter(self, node):
        self._stat_default_value(node)
        
    def _stat_default_value(self, node):
        if hasattr(node, 'default') and type(node.default) == type('') and \
           node.default[0] in ('[' , '{'):
            self.data['bad_default_value'].append(node.fullname())
        
    def _stat(self, node):
        """
        standard stats for function, class and module objects
        """
        data = self.data
        data['stat'].setdefault(node.TYPE, [0, 0])
        data['stat'][node.TYPE][0] += 1
        if node.doc:
            data['stat'][node.TYPE][1] += 1
        else:
            if node.name == '__init__':
                s = '%s.%s'%(node.fullname(), node.name)
            else:
                s = node.fullname()
            self.data['undocummented'].append(node)
        for mod, asmod in node.imports:
            if asmod == '*':
                name = node.fullname()
                data['import*'].setdefault(name, 0)
                data['import*'][name] += 1
            if not isinstance(mod, AbstractBase):
                if not is_standard_module(mod.split('.')[0]):
                    data['unresolved'].setdefault(mod, []).append(node.fullname())

        
# print stats as text ##########################################################
def print_stats(node, data, verbose=0):
    """
    nicely format statistics results
    """
    print '-'*80
    print '                  STATISTICS FOR %s %s'%(node.TYPE.upper(),node.name)
    _separ()
    print 'A few numbers:\n'
    print ' %d modules \t (<%d / %d%% documented)' % data['nice_stats']['module']
    if data['nice_stats']['class'][0] == 0:
        print '    No classes defined.'
    else:
        print ' %d classes \t (<%d / %d%% documented)' % data['nice_stats']['class']
    if data['nice_stats']['function'][0] == 0:
        print '    No functions defined.'
    else:
        print ' %d functions \t (<%d / %d%% documented)' % data['nice_stats']['function']
    if verbose:
        for n in data['undocummented']: 
            print '%s %s is not documented'%(n.TYPE, n.fullname())
    print 
    _separ()    
    tree_defs = make_tree_defs(data['unresolved'].items())
    if tree_defs:
        print 'Unresolved external imports:'
        print repr_tree_defs(tree_defs)
    else:
        print 'No unresolved external imports 8-)\n'
    _separ()
    if data['import*']:
        print 'Bad guys which import *:\n'
        for file, num in data['import*'].items():
            print ' %s (%s)'% (file, num)
        print
    else:
        print 'Good, no files do "from F import * " !!\n'
    _separ()
    if data['cycle']:
        print 'Cyclic dependencies:\n'
        for _from, to in data['cycle']:
            print ' from %s to %s'% (_from.name, to.name)
        print
    else:
        print 'Good, No cyclic dependencies\n'
    _separ()
    if data['bad_default_value']:
        print "Are you sure you want a list/dictionnary default value for \
the following\nfunctions parameters ?\n"
        for param in data['bad_default_value']:
            print ' %s'% (param)
    else:
        print 'Good, no dangerous default value ([] or {})\n'
    print
    print '-'*80

def _separ():
    """
    print a separator
    """
    print '%s%s\n'%(' '*10, '-'*60)

# utilities to represents imports as tree ######################################

def make_tree_defs(mod_files_list):
    tree_defs = {}
    for mod, files in mod_files_list:
        prefixes = mod.split('.')
        d = (tree_defs, ())
        for prefix in prefixes:
            d = d[0].setdefault(prefix, [{}, []])
        d[1] += files
    return tree_defs

def repr_tree_defs(data, indent_str=None):
    """ return a string which represents imports as a tree"""
    s = ''
    l = data.items()
    for i in range(len(l)):
        mod, (sub, files) = l[i]
        if not files: files = ''
        if indent_str is None:
            s = '%s\n%s %s\n' % (s, mod, files)
            sub_indent_str = '  '
        else:
            s = '%s%s\-%s %s\n' % (s, indent_str, mod, files)
            if i == len(l)-1:
                sub_indent_str = '%s  ' % indent_str
            else: sub_indent_str = '%s| ' % indent_str
        s = '%s%s'% (s, repr_tree_defs(sub, sub_indent_str))
    return s


